/*****************************************************************************
 *                                                                           *
 * The breve Simulation Environment                                          *
 * Copyright (C) 2000, 2001, 2002, 2003 Jonathan Klein                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License as published by      *
 * the Free Software Foundation; either version 2 of the License, or         *
 * (at your option) any later version.                                       *
 *                                                                           *
 * This program is distributed in the hope that it will be useful,           *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
 * GNU General Public License for more details.                              *
 *                                                                           *
 * You should have received a copy of the GNU General Public License         *
 * along with this program; if not, write to the Free Software               *
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
	the breve simulation environment plugin API version 1.7

	Documentation on using the breve plugin API is included in
	the documentation distributed with breve.
*/

typedef struct slVector slVector;
typedef struct stEval stEval;
typedef struct stEvalListHead stEvalListHead;
typedef struct stEvalList stEvalList;
typedef struct stEvalHash stEvalHash;
typedef struct stData stData;
typedef struct stInstance stInstance;

/* breve output error codes */

#define EC_ERROR -1		/* causes the simulation to terminate */
#define EC_OK 1			/* normal output--simulation continues */

/* these entries appear in the stEval field */ 


enum atomicTypes {
	AT_UNDEFINED = 0,
	AT_NULL,
	AT_INT,
    AT_DOUBLE,
	AT_STRING,
	AT_OBJECT,
	AT_POINTER,
	AT_VECTOR,
	AT_MATRIX,
	AT_LIST,
	AT_ARRAY,
	AT_DATA,
	AT_HASH,
	AT_TYPE
};

struct slVector {
	double x;
	double y;
	double z;
};

struct stEval {
	unsigned char type;
	unsigned char retain;

	union {
		double doubleValue;
		int intValue;
		slVector vectorValue;
		double matrixValue[3][3];
		void *pointerValue;
		char *stringValue;
		stEvalHash *hashValue;
		stData *dataValue;
		stInstance *instanceValue;
		stEvalListHead *listValue;
	} values;
};

/*
	lists in steve are held internally by the stEvalListHead which
	holds a doubly-linked list of stEvalList structures.  

	do not modify this structure.
*/

struct stEvalListHead {
	int count;
	int retainCount; 
	stEvalList *start;
	stEvalList *end;

    stEvalList *lastAccess;
    int lastAccessIndex;
};

struct stEvalList { 
	int number;
 
	double sortValue;
 
	stEval eval;
 
	stEvalList *next;
	stEvalList *previous;
};

/*
 * the stData struct and associated functions.
 */

struct stData {
    void *data;
    int length;
    int retainCount;
};

stData *stNewData(void *data, int length);
void stFreeData(stData *data);

char *slStrdup(char *string);

/* 
	the following functions are used to create & edit evaluation lists.
	you can create stEvalLists to be returned by your functions.  don't
	worry about freeing stEvalLists, this is done by the engine.
*/

stEvalListHead *stNewEvalList();
stEval *stEvalListPrepend(stEvalListHead *a, stEval *eval);
stEval *stEvalListAppend(stEvalListHead *a, stEval *eval);

/* use these macros to treat stEval pointers like specific types */

#define STINT(e)        ((e)->values.intValue)
#define STFLOAT(e)      ((e)->values.doubleValue)
#define STDOUBLE(e)     ((e)->values.doubleValue)
#define STSTRING(e)     ((e)->values.stringValue)
#define STVECTOR(e)     ((e)->values.vectorValue)
#define STMATRIX(e)     ((e)->values.matrixValue)
#define STOBJECT(e)     ((e)->values.instanceValue)
#define STPOINTER(e)    ((e)->values.pointerValue)
#define STDATA(e)       ((e)->values.dataValue)
#define STHASH(e)       ((e)->values.hashValue)
#define STLIST(e)       ((e)->values.listValue)

int stNewSteveCall(void *n, char *name, int (*call)(stEval *argumentArray, stEval *returnValue, void *callingInstance), int rtype, ...);

/*
	to call breve methods from your plugin, use the following method.  
	the arguments to the method should be placed in an array of stEval
	structs in the order that they appear in the methods definition
	(keywords are not used with this interface).  you must also supply
	the number of arguments supplied.

	the stEval pointed to in result will return the return value (if 
	any) of the method.

	the return value of the function itself is either EC_OK or EC_ERROR
	depending on whether an error occurred while calling the specified
	method.
*/

int stCallMethodByNameWithArgs(void *instance, char *name, stEval **args, int argcount, stEval *result);

/*
	to print error messages to the breve console, use the function slMessage.
	the first argument should always be DEBUG_ALL.  the second argument is a 
	"printf" style format string, and subsequent arguments correspond to the 
	arguments referenced in the format string.
*/

#define DEBUG_ALL	0

int slMessage(int level, char *format, ...);

/*
	slMalloc and slFree are plugin replacements for malloc and free which 
	do some simple memory management.  Their use is optional, but if you 
	choose to use one, then you must use the other as well.  An slMalloc
	call must be balanced by an slFree.
*/

void *slMalloc(int size);
void *slRealloc(void *pointer, int size);
void slFree(void *pointer);

/*
	pass in the callingInstance pointer, and this function returns 
	a FILE* to which fprintf output can be directed in order to send 
	it to the breve log.
*/

FILE *slGetLogFilePointer(void *callingInstance);

char *stPluginFindFile(char *file, void *i);

#ifdef __cplusplus
}
#endif /* __cplusplus */
